# Autogenerated by Astropy-affiliated package astroquery_integral's setup.py on 2021-03-20 17:08:29 UTC
import datetime

version = "0.1.1"
githash = "cfe9861428b7dde59a73038285ee72895b54766f"


major = 0
minor = 1
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2021, 3, 20, 17, 8, 29)
debug = False

astropy_helpers_version = "4.0.1"
